@@ -76,11 +76,14 @@
 #include <string.h>
 #include <assert.h>
 #include <ctype.h>
+#include <inttypes.h>
+#include <stdbool.h>
 
 #include "jasper/jas_math.h"
 #include "jasper/jas_image.h"
 #include "jasper/jas_malloc.h"
 #include "jasper/jas_string.h"
+#include "jasper/jas_debug.h"
 
 /******************************************************************************\
 * Types.
@@ -303,15 +306,16 @@ void jas_image_destroy(jas_image_t *image)
 	jas_free(image);
 }
 
-static jas_image_cmpt_t *jas_image_cmpt_create(uint_fast32_t tlx, uint_fast32_t tly,
-  uint_fast32_t hstep, uint_fast32_t vstep, uint_fast32_t width, uint_fast32_t
-  height, uint_fast16_t depth, bool sgnd, uint_fast32_t inmem)
+static jas_image_cmpt_t *jas_image_cmpt_create(uint_fast32_t tlx,
+  uint_fast32_t tly, uint_fast32_t hstep, uint_fast32_t vstep,
+  uint_fast32_t width, uint_fast32_t height, uint_fast16_t depth, bool sgnd,
+  uint_fast32_t inmem)
 {
 	jas_image_cmpt_t *cmpt;
-	long size;
+	size_t size;
 
 	if (!(cmpt = jas_malloc(sizeof(jas_image_cmpt_t)))) {
-		return 0;
+		goto error;
 	}
 
 	cmpt->type_ = JAS_IMAGE_CT_UNKNOWN;
@@ -326,23 +330,35 @@ static jas_image_cmpt_t *jas_image_cmpt_create(uint_fast32_t tlx, uint_fast32_t
 	cmpt->stream_ = 0;
 	cmpt->cps_ = (depth + 7) / 8;
 
-	size = cmpt->width_ * cmpt->height_ * cmpt->cps_;
-	cmpt->stream_ = (inmem) ? jas_stream_memopen(0, size) : jas_stream_tmpfile();
+	// Compute the number of samples in the image component, while protecting
+	// against overflow.
+	// size = cmpt->width_ * cmpt->height_ * cmpt->cps_;
+	if (!jas_safe_size_mul(cmpt->width_, cmpt->height_, &size) ||
+	  !jas_safe_size_mul(size, cmpt->cps_, &size)) {
+		goto error;
+	}
+	cmpt->stream_ = (inmem) ? jas_stream_memopen(0, size) :
+	  jas_stream_tmpfile();
 	if (!cmpt->stream_) {
-		jas_image_cmpt_destroy(cmpt);
-		return 0;
+		goto error;
 	}
 
 	/* Zero the component data.  This isn't necessary, but it is
 	convenient for debugging purposes. */
+	/* Note: conversion of size - 1 to long can overflow */
 	if (jas_stream_seek(cmpt->stream_, size - 1, SEEK_SET) < 0 ||
 	  jas_stream_putc(cmpt->stream_, 0) == EOF ||
 	  jas_stream_seek(cmpt->stream_, 0, SEEK_SET) < 0) {
-		jas_image_cmpt_destroy(cmpt);
-		return 0;
+		goto error;
 	}
 
 	return cmpt;
+
+error:
+	if (cmpt) {
+		jas_image_cmpt_destroy(cmpt);
+	}
+	return 0;
 }
 
 static void jas_image_cmpt_destroy(jas_image_cmpt_t *cmpt)
@@ -640,6 +656,24 @@ int jas_image_fmtfromname(char *name)
 * Miscellaneous operations.
 \******************************************************************************/
 
+bool jas_image_cmpt_domains_same(jas_image_t *image)
+{
+	int cmptno;
+	jas_image_cmpt_t *cmpt;
+	jas_image_cmpt_t *cmpt0;
+
+	cmpt0 = image->cmpts_[0];
+	for (cmptno = 1; cmptno < image->numcmpts_; ++cmptno) {
+		cmpt = image->cmpts_[cmptno];
+		if (cmpt->tlx_ != cmpt0->tlx_ || cmpt->tly_ != cmpt0->tly_ ||
+		  cmpt->hstep_ != cmpt0->hstep_ || cmpt->vstep_ != cmpt0->vstep_ ||
+		  cmpt->width_ != cmpt0->width_ || cmpt->height_ != cmpt0->height_) {
+			return 0;
+		}
+	}
+	return 1;
+}
+
 uint_fast32_t jas_image_rawsize(jas_image_t *image)
 {
 	uint_fast32_t rawsize;
@@ -674,8 +708,9 @@ int jas_image_addcmpt(jas_image_t *image, int cmptno,
   jas_image_cmptparm_t *cmptparm)
 {
 	jas_image_cmpt_t *newcmpt;
-	if (cmptno < 0)
+	if (cmptno < 0) {
 		cmptno = image->numcmpts_;
+	}
 	assert(cmptno >= 0 && cmptno <= image->numcmpts_);
 	if (image->numcmpts_ >= image->maxcmpts_) {
 		if (jas_image_growcmpts(image, image->maxcmpts_ + 128)) {
@@ -790,7 +825,8 @@ static int jas_image_growcmpts(jas_image_t *image, int maxcmpts)
 	jas_image_cmpt_t **newcmpts;
 	int cmptno;
 
-	newcmpts = (!image->cmpts_) ? jas_alloc2(maxcmpts, sizeof(jas_image_cmpt_t *)) :
+	newcmpts = (!image->cmpts_) ? jas_alloc2(maxcmpts,
+	  sizeof(jas_image_cmpt_t *)) :
 	  jas_realloc2(image->cmpts_, maxcmpts, sizeof(jas_image_cmpt_t *));
 	if (!newcmpts) {
 		return -1;
@@ -803,8 +839,8 @@ static int jas_image_growcmpts(jas_image_t *image, int maxcmpts)
 	return 0;
 }
 
-int jas_image_copycmpt(jas_image_t *dstimage, int dstcmptno, jas_image_t *srcimage,
-  int srccmptno)
+int jas_image_copycmpt(jas_image_t *dstimage, int dstcmptno,
+  jas_image_t *srcimage, int srccmptno)
 {
 	jas_image_cmpt_t *newcmpt;
 	if (dstimage->numcmpts_ >= dstimage->maxcmpts_) {
@@ -837,7 +873,7 @@ void jas_image_dump(jas_image_t *image, FILE *out)
 	jas_image_cmpt_t *cmpt;
 	for (cmptno = 0; cmptno < image->numcmpts_; ++cmptno) {
 		cmpt = image->cmpts_[cmptno];
-		fprintf(out, "prec=%d, sgnd=%d, cmpttype=%d\n", cmpt->prec_,
+		fprintf(out, "prec=%d, sgnd=%d, cmpttype=%"PRIiFAST32"\n", cmpt->prec_,
 		  cmpt->sgnd_, cmpt->type_);
 		width = jas_image_cmptwidth(image, cmptno);
 		height = jas_image_cmptheight(image, cmptno);
@@ -1214,13 +1250,38 @@ static void jas_image_calcbbox2(jas_image_t *image, jas_image_coord_t *tlx,
 	*bry = tmpbry;
 }
 
+static inline long decode_twos_comp(ulong c, int prec)
+{
+	long result;
+	assert(prec >= 2);
+	jas_eprintf("warning: support for signed data is untested\n");
+	// NOTE: Is this correct?
+	result = (c & ((1 << (prec - 1)) - 1)) - (c & (1 << (prec - 1)));
+	return result;
+}
 
+static inline ulong encode_twos_comp(long n, int prec)
+{
+	ulong result;
+	assert(prec >= 2);
+	jas_eprintf("warning: support for signed data is untested\n");
+	// NOTE: Is this correct?
+	if (n < 0) {
+		result = -n;
+		result = (result ^ 0xffffffffUL) + 1;
+		result &= (1 << prec) - 1;
+	} else {
+		result = n;
+	}
+	return result;
+}
 
 static int getint(jas_stream_t *in, int sgnd, int prec, long *val)
 {
 	long v;
 	int n;
 	int c;
+	assert((!sgnd && prec >= 1) || (sgnd && prec >= 2));
 	n = (prec + 7) / 8;
 	v = 0;
 	while (--n >= 0) {
@@ -1230,8 +1291,7 @@ static int getint(jas_stream_t *in, int sgnd, int prec, long *val)
 	}
 	v &= ((1 << prec) - 1);
 	if (sgnd) {
-		/* XXX - Do something here. */
-		abort();
+		*val = decode_twos_comp(v, prec);
 	} else {
 		*val = v;
 	}
@@ -1242,10 +1302,13 @@ static int putint(jas_stream_t *out, int sgnd, int prec, long val)
 {
 	int n;
 	int c;
+	bool s;
+	ulong tmp;
+	assert((!sgnd && prec >= 1) || (sgnd && prec >= 2));
 	if (sgnd) {
-		/* XXX - Do something here. */
-		abort();
+		val = encode_twos_comp(val, prec);
 	}
+	assert(val >= 0);
 	val &= (1 << prec) - 1;
 	n = (prec + 7) / 8;
 	while (--n >= 0) {
@@ -1319,6 +1382,8 @@ jas_eprintf("IMAGE\n");
 jas_image_dump(image, stderr);
 #endif
 
+	outimage = 0;
+	xform = 0;
 	if (!(inimage = jas_image_copy(image)))
 		goto error;
 	image = 0;
@@ -1329,16 +1394,20 @@ jas_image_dump(image, stderr);
 		for (i = 1; i < jas_image_numcmpts(inimage); ++i) {
 			hstep = jas_image_cmpthstep(inimage, i);
 			vstep = jas_image_cmptvstep(inimage, i);
-			if (hstep < minhstep)
+			if (hstep < minhstep) {
 				minhstep = hstep;
-			if (vstep < minvstep)
+			}
+			if (vstep < minvstep) {
 				minvstep = vstep;
+			}
 		}
 		n = jas_image_numcmpts(inimage);
 		for (i = 0; i < n; ++i) {
 			cmpttype = jas_image_cmpttype(inimage, i);
-			if (jas_image_sampcmpt(inimage, i, i + 1, 0, 0, minhstep, minvstep, jas_image_cmptsgnd(inimage, i), jas_image_cmptprec(inimage, i)))
+			if (jas_image_sampcmpt(inimage, i, i + 1, 0, 0, minhstep, minvstep,
+			  jas_image_cmptsgnd(inimage, i), jas_image_cmptprec(inimage, i))) {
 				goto error;
+			}
 			jas_image_setcmpttype(inimage, i + 1, cmpttype);
 			jas_image_delcmpt(inimage, i);
 		}
@@ -1349,8 +1418,9 @@ jas_image_dump(image, stderr);
 	hstep = jas_image_cmpthstep(inimage, 0);
 	vstep = jas_image_cmptvstep(inimage, 0);
 
-	inprof = jas_image_cmprof(inimage);
-	assert(inprof);
+	if (!(inprof = jas_image_cmprof(inimage))) {
+		abort();
+	}
 	numinclrchans = jas_clrspc_numchans(jas_cmprof_clrspc(inprof));
 	numinauxchans = jas_image_numcmpts(inimage) - numinclrchans;
 	numoutclrchans = jas_clrspc_numchans(jas_cmprof_clrspc(outprof));
@@ -1358,8 +1428,9 @@ jas_image_dump(image, stderr);
 	numoutchans = numoutclrchans + numoutauxchans;
 	prec = 8;
 
-	if (!(outimage = jas_image_create0()))
+	if (!(outimage = jas_image_create0())) {
 		goto error;
+	}
 
 	/* Create a component for each of the colorants. */
 	for (i = 0; i < numoutclrchans; ++i) {
@@ -1392,18 +1463,22 @@ jas_image_dump(image, stderr);
 	tmpprof = 0;
 	jas_image_setclrspc(outimage, jas_cmprof_clrspc(outprof));
 
-	if (!(xform = jas_cmxform_create(inprof, outprof, 0, JAS_CMXFORM_OP_FWD, intent, 0)))
+	if (!(xform = jas_cmxform_create(inprof, outprof, 0, JAS_CMXFORM_OP_FWD,
+	  intent, 0))) {
 		goto error;
+	}
 
 	inpixmap.numcmpts = numinclrchans;
-	incmptfmts = malloc(numinclrchans * sizeof(jas_cmcmptfmt_t));
-	assert(incmptfmts);
+	if (!(incmptfmts = jas_alloc2(numinclrchans, sizeof(jas_cmcmptfmt_t)))) {
+		abort();
+	}
 	inpixmap.cmptfmts = incmptfmts;
 	for (i = 0; i < numinclrchans; ++i) {
 		j = jas_image_getcmptbytype(inimage, JAS_IMAGE_CT_COLOR(i));
 		assert(j >= 0);
-		if (!(incmptfmts[i].buf = malloc(width * sizeof(long))))
+		if (!(incmptfmts[i].buf = jas_alloc2(width, sizeof(long)))) {
 			goto error;
+		}
 		incmptfmts[i].prec = jas_image_cmptprec(inimage, j);
 		incmptfmts[i].sgnd = jas_image_cmptsgnd(inimage, j);
 		incmptfmts[i].width = width;
@@ -1411,14 +1486,15 @@ jas_image_dump(image, stderr);
 	}
 
 	outpixmap.numcmpts = numoutclrchans;
-	outcmptfmts = malloc(numoutclrchans * sizeof(jas_cmcmptfmt_t));
-	assert(outcmptfmts);
+	if (!(outcmptfmts = jas_alloc2(numoutclrchans, sizeof(jas_cmcmptfmt_t)))) {
+		abort();
+	}
 	outpixmap.cmptfmts = outcmptfmts;
 
 	for (i = 0; i < numoutclrchans; ++i) {
 		j = jas_image_getcmptbytype(outimage, JAS_IMAGE_CT_COLOR(i));
 		assert(j >= 0);
-		if (!(outcmptfmts[i].buf = malloc(width * sizeof(long))))
+		if (!(outcmptfmts[i].buf = jas_alloc2(width, sizeof(long))))
 			goto error;
 		outcmptfmts[i].prec = jas_image_cmptprec(outimage, j);
 		outcmptfmts[i].sgnd = jas_image_cmptsgnd(outimage, j);
@@ -1429,22 +1505,26 @@ jas_image_dump(image, stderr);
 	for (i = 0; i < height; ++i) {
 		for (j = 0; j < numinclrchans; ++j) {
 			k = jas_image_getcmptbytype(inimage, JAS_IMAGE_CT_COLOR(j));
-			if (jas_image_readcmpt2(inimage, k, 0, i, width, 1, incmptfmts[j].buf))
+			if (jas_image_readcmpt2(inimage, k, 0, i, width, 1,
+			  incmptfmts[j].buf))
 				goto error;
 		}
 		jas_cmxform_apply(xform, &inpixmap, &outpixmap);
 		for (j = 0; j < numoutclrchans; ++j) {
 			k = jas_image_getcmptbytype(outimage, JAS_IMAGE_CT_COLOR(j));
-			if (jas_image_writecmpt2(outimage, k, 0, i, width, 1, outcmptfmts[j].buf))
+			if (jas_image_writecmpt2(outimage, k, 0, i, width, 1,
+			  outcmptfmts[j].buf))
 				goto error;
 		}
 	}
 
-	for (i = 0; i < numoutclrchans; ++i)
+	for (i = 0; i < numoutclrchans; ++i) {
 		jas_free(outcmptfmts[i].buf);
+	}
 	jas_free(outcmptfmts);
-	for (i = 0; i < numinclrchans; ++i)
+	for (i = 0; i < numinclrchans; ++i) {
 		jas_free(incmptfmts[i].buf);
+	}
 	jas_free(incmptfmts);
 	jas_cmxform_destroy(xform);
 	jas_image_destroy(inimage);
